/*!
 * \file dm_test.cc
 * \brief Dynamic memory manager test
 * \author Petr Gladkikh, 2001, 2002
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *     
 */

#include <iostream>
#include <unistd.h>
#include <typeinfo>

#include "mm/dm.h"
#include "debug/tester.h"

using namespace std;
using namespace dm;

#undef NDEBUG 

const dm::uint POOL_SIZE = 10*1024*1024; 
MemoryPool mm(POOL_SIZE);

const dm::uint BIG = 1000;


void printDMStatistics(const MemoryPool& dmm)
{
    cout << "pool bottom = " << dmm.getPoolBottom() << endl;
    cout << "sbrk(0) = " << sbrk(0) << endl;
    cout << "capacity = " << dmm.getCapacity() << endl;
    cout << "free space amount = " << dmm.getFreeAmount() << endl;
    cout << "block header size = " << sizeof(Block) << endl;

    MemoryPool::Statistics stat(&dmm);
    cout << "number of pointers = " << stat.pointers.pointers_number << endl;
    cout << "mask capacity = " << stat.pointers.capacity << endl;
    cout << "pointer mask fill = " << (double)stat.pointers.pointers_number/stat.pointers.capacity
	 << endl;
    cout << "pointers_in_stack = " << stat.pointers_in_stack << endl;
}


// --------------------------------------------------------
// Tests


//  ----- group a - scalar values -----


bool test_a_1()
{
    dynamic<int> a(0);
    assert(a == 0);
    a.init();
    assert(a != 0);
    *a = 1;
    *a = *a + 10;
    assert(*a == 11);
    a.remove();
    assert(a.ptr() == 0);
    return mm.selfTest(cout);
}


bool test_a_2()
{
    dynamic<int> b(1);
    assert(b != 0);
    *b = 0;
    assert(*b == 0);
    b.remove();
    assert(b == 0);
    return mm.selfTest(cout);
}


bool test_a_3()
{
    dynamic<float> a(1);
    dynamic<float> b(1);
    assert(a != 0);
    assert(b != 0);
    assert(*b == 0);
    assert(*a == 0);
    assert(a.ptr() != b.ptr());
    *a += 1.1;
    *b = 3;
    *b += *a;
    assert(*a == (float)1.1);
    assert(*b == (float)4.1);
    return mm.selfTest(cout);
}


//  ----- group b - references -----


bool test_b_1()
{
    dynamic<int> a(1), b(1);
    assert(a.ptr() != b.ptr());
    *a = 1;
    *b = 3;
    assert(*a != *b);
    a = b;
    assert(*a == *b);
    assert(a == b);
    *a = 44;
    assert(*b == 44);
    *b = 33;
    assert(*b == 33);
    assert(*a == *b);
    return mm.selfTest(cout);
}


bool test_b_2()
{
    dynamic<int> a(1), b(1);
    a[0] = 1;
    assert(a[0] == *a);
    b = a;
    assert(*a == b[0]);
    a.init();
    assert(*a == 0);
    assert(*a != *b);
    assert(*b == 1);
    return mm.selfTest(cout);
}


bool test_b_3()
{
    dynamic<int> a(1);
    MemoryPool::Statistics stat0(&mm);
    for (unsigned i = 0; i < BIG; i++)
    {
	a.init();
    }
    MemoryPool::Statistics stat1(&mm);
    assert(stat0.free_space == stat1.free_space);
    return mm.selfTest(cout);
}


class Chain
{
    public:
	unsigned key;
	dynamic<Chain> next;
};


bool test_b_4()
{
    // try to build list
    const unsigned length = BIG;
    dm::uint free_space_before = mm.getFreeAmount();
    dynamic<Chain> chain(1);
    dynamic<Chain> head(0);
    {
	head = chain;
	assert(mm.isCorrectPointer(chain.ptr()));
	assert(mm.isCorrectPointer(head.ptr()));
	for (unsigned i = 0; i < length; i++)
	{
	    chain->next.init();
	    chain->key = i;
	    assert(chain != chain->next);
	    chain = chain->next;
	}
    }
    {
	// now cleanup it
	chain = head;
	head.remove();
	for (unsigned i = 0; i < length; i++)
	{
	    dynamic<Chain> prev = chain;
	    assert(chain != chain->next);
	    assert(chain->key == i);
	    ///assert(mm.isCorrectPointer(chain.ptr()));
	    ///assert(mm.isCorrectPointer(head.ptr()));
	    ///assert(mm.isCorrectPointer(chain->next.ptr()));
	    chain = chain->next;	
	}
    }
    chain.remove();
    // mm.dump(cout); // DEBUG
    dm::uint free_space_after = mm.getFreeAmount();
    // cout << free_space_before << "--->" << free_space_after << endl; // DEBUG
    assert(free_space_before == free_space_after);
    return mm.selfTest(cout);
}


bool test_b_5()
{
    dm::uint free_space_before = mm.getFreeAmount();
    dynamic<Chain> chain(1);

    for (unsigned i = 0; i < BIG; i++)
    {
	chain->next.init();
	chain = chain->next;
    }
    chain.remove();
    dm::uint free_space_after = mm.getFreeAmount();
    assert(free_space_before == free_space_after);
    return mm.selfTest(cout);
}


bool test_b_6()
{
    dm::uint free_space_before = mm.getFreeAmount();
    dynamic<Chain> chain(1);
    dynamic<Chain> handle(0);

    handle = chain;
    for (unsigned i = 0; i < 10; i++)
    {
	chain->next.init();
	chain = chain->next;
    }
    chain->next = handle;
    chain.remove();
    handle.remove();
    // now we have cycle in memory without references from stack
    mm.optimize();
    dm::uint free_space_after = mm.getFreeAmount();
    assert(free_space_before == free_space_after);
    return mm.selfTest(cout);
}


bool test_b_7()
{
    dm::uint free_space_before = mm.getFreeAmount();
    const unsigned NOF_FANS = 200;
    {
	dynamic< dynamic<int> > fans(NOF_FANS);
	
	fans[0].init();
	*fans[0] = 0;
	for (unsigned i = 0; i < NOF_FANS; i++)
	{
	    fans[i] = fans[0];
	    *fans[i] = i;
	    assert(fans[0] == fans[i]);
	}
	assert(free_space_before > mm.getFreeAmount());
	for (unsigned i = 0; i < NOF_FANS; i++)
	{
	    fans[i].remove();
	}
    }
    dm::uint free_space_after = mm.getFreeAmount();
    assert(free_space_before == free_space_after);
    return mm.selfTest(cout);
}


//  ----- group c - resize -----


bool test_c_1()
{
    dm::uint free_space_before = mm.getFreeAmount();
    const int CHUNK_SIZE = 500;
    dynamic<int> Z(CHUNK_SIZE);
    for (int i = 0; i < CHUNK_SIZE; i++)
    {
	Z[i] = CHUNK_SIZE - i;
    }

    // Z.resize(CHUNK_SIZE + 33); // TODO resize not implemented yet
   
    for (int i = 0; i < CHUNK_SIZE; i++)
    {
	assert(Z[i] == CHUNK_SIZE - i);
    }
    
    dm::uint free_space_after = mm.getFreeAmount();
    assert(free_space_before == free_space_after);
    return mm.selfTest(cout);
}


//  ----- group d - defrag -----


bool test_d_1()
{
    const unsigned CHUNK_SIZE = 5000;
    dynamic<Chain> Z(1);
    dynamic<Chain> X(1);
    dynamic<Chain> z = Z;
    dynamic<Chain> x = X;

    for (unsigned i = 0; i < CHUNK_SIZE; i++)
    {
	z->key = i;
	z->next.init();
	z = z->next;

	x->key = i;
	x->next.init();
	x = x->next;
    }
    
    x = X;
    for (unsigned i = 0; i < CHUNK_SIZE; i++)
    {
	assert(x->key == i);
	x = x->next;
    }
    
    X.remove();
    dm::uint free_space_before = mm.getFreeAmount();
    mm.optimize(); 
    dm::uint free_space_after = mm.getFreeAmount();
    // free blocks shold be merged
    // cout << free_space_before << "--->" << free_space_after << endl; // DEBUG
    assert(free_space_before < free_space_after);
    // cout << "[STAT]" << endl; printDMStatistics(mm); cout.flush(); // DEBUG
   
    z = Z;
    for (unsigned i = 0; i < CHUNK_SIZE; i++)
    {
	assert(z->key == i);
	z = z->next;
    }
   
    return mm.selfTest(cout);
}


//  ----------------------------


TestCase tests[] =
{
    { test_a_1, "a1" },
    { test_a_2, "a2" },
    { test_a_3, "a3" },
    { test_b_1, "b1" },
    { test_b_2, "b2" },
    { test_b_3, "b3" },
    { test_b_4, "b4" },
    { test_b_5, "b5" }, 
    { test_b_6, "b6" },
    { test_b_7, "b7" },
    { test_c_1, ""   },
    { test_d_1, "d1" },
    { 0,        ""   }  /* nil function pointer is a stop token */
};


// end tests list
// -------------------------------------------------------- 


int main(int argc, char** argv)
{
    cout << "Dynamic memory manager (dm) test." << endl;
    printDMStatistics(mm);

    if (!mm.selfTest(cout))
    {
	cout << "Post initialization self test FAILED." << endl; 
	return 0;
    };

    cout << endl;
    if (runTests(tests))
    {
	cout << "TESTS PASSED" << endl;
	return 0;
    }
    else
    {
	cout << endl;
	printDMStatistics(mm);
	cout << "TESTS FAILED" << endl;
	return 1;
    };
}

