/*!
   \file dm.cc
   \brief Dynamic memory manager core implementation. 
   \author (c) Petr Gladkikh, 2001, 2002 (Petr.Gladkikh@bacup.ru).
   $Id: dm.cc,v 1.6 2002/10/20 15:19:45 petr Exp $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Library General Public License as published
    by the Free Software Foundation; either version 2, or (at your option) any
    later version.
 */

#include "dm.h"
#include <cassert>
#include <iostream>
#include <algorithm>

namespace dm 
{
    
    // ---------------------------------------------------------
    // class Reference 

    void Reference::dump(std::ostream& stream,
			 void** to,
			 const std::string& prefix,
			 const std::string& name)
    {
	stream << prefix << (name.empty() ? getClassName() : name) << "::";
	stream << to << ":";
	if (to != 0)
	    stream << "*ref=" << *to << ":";
	stream << std::endl;
    };


    // ---------------------------------------------------------
    // class GenericDynamic


    MemoryPool* GenericDynamic::memory_pool = 0;


    void GenericDynamic::setPool(MemoryPool* const newMemoryPool)
    {
	memory_pool = newMemoryPool;
    };


    // ---------------------------------------------------------
    // class dynamic
    //

    // (all functions are inline now)


    // ---------------------------------------------------------
    // class Block 
    //

    MemoryPool* Block::memory_pool = 0;


    void Block::setPool(MemoryPool* const newMemoryPool)
    {
	memory_pool = newMemoryPool;
    };


    /*!
       Bootstrap and explicit constructor. 
       Use constructor with "parent" parameter whenever possible.
       \note Should be called using "new ptr ctor;" construction only.
     */
    Block::Block(const uint new_size,
		 Block* const new_prev,
		 Block* const new_next)
	: prev(new_prev),
	  next(new_next), 
	  size(new_size)
    {
	assert(new_size > sizeof(*this));
	flags.invalid = true;
	flags.used = false;
	initReferences();
	flags.invalid = false;
    };


    Block::Block(Block& parent)
	: prev(parent.prev),
	  next(parent.next), 
	  size(parent.size)
    {
	flags.invalid = true;
	flags.under_destruction = false; 
	initReferences();
	flags.used = false;
	if (&parent != this)
	{
	    // perform complete initialization
	    size = (const char*)(&parent) + parent.size - (const char*)(this);
	    assert(0 < size);
	    prev = &parent;
	    next = parent.getNext(); 
	    parent.next = this; 
	    if (next != 0)
	    {
		next->prev = this;
	    }
	    assert( (next!=0 && prev!=0 && prev<next)  ||  size>0 );
	    assert( (next!=0 && this<next)  ||  (next==0 && size>0) );
	    parent.size -= size;
	}
	else
	{
	    // inherit field values from parent
	}
	flags.invalid = false;
    };


    Block* Block::getNewInstanceAddr(Block* const parent,
				     const uint new_size)
    {
	assert(parent != 0);
	assert(new_size <= parent-> getSize());
	Block* newBlockAddr = (Block*)((char*)(parent) + parent->getSize() - new_size);
	assert(parent <= newBlockAddr);
	if (static_cast<long int>
	    ((const char*)(newBlockAddr) - (const char*)(parent + 1)
	    ) <= CONSUME_TRESHOLD
	   )
	{
	    newBlockAddr = parent;
	}
	else
	{
	    assert(parent->addressInsideBody(newBlockAddr));
	    assert(parent->addressInsideBody((char* const)(newBlockAddr)
					     + new_size - 1) );
	}
	return newBlockAddr;
    };


    Block* Block::initInstance(Block* const parent,
			       uint new_size)    
    {
	return t_initInstance<Block>(parent, new_size);
    };


    void Block::initReferences()
    {
	reference_counter = 0;
    }


    /*!
        \par Warning
	If destination is larger
	then remaining area rest part stays uninitialized.	
    */    
    void Block::operator= (const Block& another_block)
    {
	memcpy(getBody(), another_block.getBody(),
		std::min(getCapacity(), another_block.getCapacity()) );
    };


    /*!
	\todo Enhance this : perhaps next block is free and we can use
	it or we can move next block or we can split current block if
	older one was too big.
    */
    Block* Block::resize(uint new_size)
    {
	assert(new_size >= getHeaderSize());
	Block*  new_self = memory_pool->allocBlock(new_size + getHeaderSize(),
						   getInitializator());
	if (new_self != 0)
	{
	    *new_self = *this;
	    this->~Block();
	}
	return new_self;
    };


    bool Block::selfTest(std::ostream& stream) const
    {
	assert(!isInvalid());
	if (getNext() != 0)
	{
	    assert(this < getNext());
	    assert(getNext()->getPrev() == this);
	    assert(reinterpret_cast<const char*>(this)
		    + sizeof(*this)
		    + getCapacity() == reinterpret_cast<const char*>(getNext()));
	}
	if (getPrev() != 0)
	{
	    assert(getPrev() < this);
	    assert(getPrev()->getNext() == this);

	    assert(reinterpret_cast<const char*>(this) - getPrev()->getSize()
		    == reinterpret_cast<const char*>(getPrev())
		  );
	}
	return true;
    };


    /*!
	\todo Check if there are pointers in the block and cleanup them.
     */
    Block::~Block()
    {
	assert(!isInvalid());
	if (isFree())
	{
	    return;
	}
	if (flags.under_destruction)
	{
	    // this destructor can eventually recursively call itself 
	    return;
	}
	flags.under_destruction = true;
	/// <<<<< forgetPointers(); >>>>> 
	markAsFree();
	if (getNext() != 0)
	    if (getNext()->isFree())
	    {
		mergeWithNext();
	    }
	bool is_consumed = false;
	if (getPrev() != 0)
	    if (getPrev()->isFree())
	    {
		getPrev()->mergeWithNext();
		is_consumed = true;
	    }
	if (! is_consumed)
	{
	    new ((void*)this) Block(*this);
	    // reincarnate self as a usual Block (but I want to be special, so very special, but ...)
	}
	else
	{
	    markAsInvalid();
	}
    }


    uint Block::getCapacity() const
    {
	return getSize() - sizeof(Block);
    };


    uint Block::getHeaderSize() const
    {
	return sizeof(Block);
    };


    /*!
       Split this block into two new blocks (if nessecary)
       and made last one available to user
     */
    Block* Block::giveBirth(const uint sizeToUse,
			    InitializationFunction dad)
    {
	assert(isFree());
	assert(getSize() >= sizeToUse);
	Block* newBlock = dad(this, sizeToUse);
	newBlock->markAsUsed();
	memory_pool-> on_block_new(newBlock);
	return newBlock;
    }
	

    /*!
       Merge this block with next one
       and make it available for allocations
     */
    void Block::mergeWithNext()
    {
	assert(isFree() && getNext() != 0);
	memory_pool-> on_block_delete(getNext());
	assert(getNext()->isFree());
	Block* oldBlock = getNext();
	if (oldBlock->getNext() != 0)
	{
	    oldBlock->getNext()->prev = this;
	}
	prev = getPrev();
	next = getNext()->getNext();
	size = getSize() + oldBlock->getSize();    
	oldBlock->markAsInvalid();
    }


    /*!
       Calculate how much bytes we can move this block down
     */
    int Block::shift_down_offset() const
    {
	assert(!isInvalid() && !isFree());
	if (getPrev() == 0 || !getPrev()->isFree()) 
	    return 0; // nothing to do
       
	// getSize() and real distance may contradict
	// TODO fix this problem
	/// uint offset = this->getPrev()->getSize(); 	   
	
	return (char*)(this->getPrev()) - (char*)this ;
    }

    /*!
       Swap this block and previous empty one.
     */
    uint Block::shift_down()
    {
	assert(!isInvalid() && !isFree());
	int offset = shift_down_offset();

	if (offset == 0) 
	    return 0; // nothing to do
       
	uint save_size = getSize(); 
	Block* prev_ptr = getPrev();
	Block* prev_prev_ptr = prev_ptr->getPrev();
	Block* next_ptr = getNext();
	Block* new_empty = (Block*)((char*)prev_ptr + save_size);
	memmove(prev_ptr, this, save_size);
	assert(!prev_ptr->isInvalid());
	prev_ptr->prev = prev_prev_ptr;
	prev_ptr->next = new_empty;
	assert(offset < 0);
	new ((void*)new_empty) Block(-offset, prev_ptr, next_ptr);
	new_empty->prev = prev_ptr;
	if (next_ptr != 0)
	{
	    next_ptr->prev = new_empty;
	    if (next_ptr->isFree())
	    {
		new_empty->mergeWithNext();
	    }
	}
	return offset;
    };


    bool Block::registerPointer(void** value)
    {
	assert(flags.used && !flags.invalid);
	assert(dynamic_cast<Block*>(this));
	assert(*value == this || *value == this + 1);
	reference_counter++;
	return true;
    }


    bool Block::forgetPointer(void** value)
    {
	assert(flags.used && !flags.invalid);
	assert(dynamic_cast<Block*>(this));
	assert(*value == this || *value == this + 1);
	
	if (value == 0 || reference_counter <= 0)
	    return false;
       
	reference_counter--;
	if (notReferenced())
	{
	    // no one wish to see me :-(
	    this->~Block(); // but ... see destructor
	}
	return true;
    };


    /*!
       Get back_references number  
     */  
    uint Block::getBRCount() const
    {
	assert(!isInvalid() && !isFree());
	return reference_counter;
    };


    bool Block::notReferenced() const
    {
	assert(!flags.invalid);
	return getReferenceCount() == 0;
    };


    uint Block::getReferenceCount() const
    {
	assert(!flags.invalid);
	return reference_counter;
    };


    void Block::dump(std::ostream& stream,
		     const std::string& prefix,
		     const std::string& name)
    {
	stream << prefix << (name.empty() ? getClassName() : name) << "::";
	stream.width(10);
	stream << this << ":";
	stream.width(8);
	stream << getSize() << ":";
	stream.width(10);
	stream << getPrev() << ":";
	stream.width(10);
	stream << getNext() << ":";
	stream.width(0);
	stream << (flags.used ? "U" : "F") 
	       << (flags.invalid ? "I" : "V") << ":"
	       << "tag=" << flags.tag << ":";
	stream << "refs=" << reference_counter << ":";
	stream << "typeid=" << typeid(*this).name() << ":";
	stream << std::endl;
    };


    // ---------------------------------------------------------
    // MemoryPool
    //

    MemoryPool::MemoryPool(const uint initialCapacity) :
	_initialized(false),
	blocks(0),
	blocks_end(0),
	reserve(0), 
	capacity(0),
	freeBlockHint(0),
	block_count(0),
	ptr_count(0)
    {
	if (initialCapacity > 0)
	{
	    Parameters parameters;
	    parameters.capacity = initialCapacity;
	    init(parameters);

	    GenericDynamic::setPool(this);
	    Block::setPool(this);
	}
    };


    MemoryPool::~MemoryPool()
    {
	if (_initialized)
	{
	    if (blocks != 0)
	    {
		::free((void*)blocks);
		blocks = 0;
	    }
	    _initialized = false;
	}
    };


    bool MemoryPool::init(const Parameters parameters)
    {
	assert(!_initialized);
	if (parameters.capacity < sizeof(Block)) return false;
	uint block_number = (parameters.capacity / sizeof(Block)) + 1;
	uint new_capacity = block_number * sizeof(Block);
	blocks = reinterpret_cast<Block*>(::malloc(new_capacity));
	if (blocks == 0) return false;
	blocks_end = blocks;
	capacity = new_capacity;
       
	pool_top = (void**)((char*)blocks + capacity);
	available_top = pointer_mask.init((void**)blocks, pool_top);
	capacity = (char*)available_top - (char*)blocks;
	pointer_stack = available_top;
	pointer_stack_top = available_top - 1;
       
	new ((void*)blocks) Block(capacity);
	block_count = 1;
       
	update_reserve();
	_initialized = true;
	return blocks != 0;
    };


    uint MemoryPool::calculate_reserve() const
    {
	return (block_count > ptr_count) ? block_count : ptr_count;
    }



    /*!
       Update memory reserve size in according to current pool state.
       
       \todo Check here if reserve can be fit into pool.
       \todo Refine reservation rules
     */
    void MemoryPool::update_reserve()
    {
	reserve = calculate_reserve();
	available_top = pointer_stack - reserve;
    }


    bool MemoryPool::optimize()
    {
	assert(_initialized);
	collectGarbage();
	defragment();
	assert(selfTest(std::cerr));
	return false;
    };


    /*!
	Mark blocks which immediately referred from known necessary blocks.
     */
    void MemoryPool::GC_extendBlockMark(Block* const block, 
					       Block** &begin,
					       Block** &end,
					       const uint tag_necessary)
    {
	assert(block != 0 && block->selfTest(std::cerr));
	
	PointerMask::iterator p(&pointer_mask, block->getBody());
	assert(block->getTag() == tag_necessary);
	Block* p_owner = 0;
	Block** put_p = begin;
	while (p.more() && (char*)p.getAdr() < ((char*)block + block->getSize()))
	{
	    p_owner = GC_findParentBlock(*(p.getAdr()));
	    if (p_owner != 0
		&& p_owner->getTag() != tag_necessary)
	    {
		{ // replace marks
		    uint pattern_tag = p_owner->getTag();
		    Block** b = begin; 
		    while (b < end)
		    {
			if ((*b)->flags.tag == pattern_tag)
			{
			    (*b)->flags.tag = tag_necessary;
			    std::swap(*put_p, *b);
			    ++put_p;
			}
			++b;
		    }
		}
	    }
	    ++p;
	};
	begin = put_p;
    }


    /*!
       Find block, on which pointer points to.
     */  
    Block* MemoryPool::GC_findParentBlock(const void* pointer)
    {
	if (pointer < blocks || available_top < pointer)
	    return 0;
       
	Block* result = (Block*)pointer - 1;
	assert(result->selfTest(std::cerr)); // check pointer: no guarentees but just in case
	return result;
    /*  
	// most roubst and slowest algorithm
	// faster one coud be "(Block*)pointer - 1" It also requires restriction on
	// pointer's value set 
	Block* block = blocks;
	while(block != 0)
	{
	    if (block <= pointer && pointer < (char*)block + block->getSize())
		return block;

	    block = block->getNext();
	}
	return 0;
    */    
    }


    /*!
       Set initial marks for blocks - prepare them for pointers trace algorithm 
     */   
    void MemoryPool::GC_setInitialMarks( Block* const * begin, Block* const * end,
						const unsigned tag_necessary,
						const unsigned tag_top)
    {
	{ 
	    Block* const * block = begin;
	    while (block < end)
	    {
		(*block)->setTag(tag_necessary+1);
		++block;
	    }
	}
	{ // consider stack pointers
	    void** pointer = pointer_stack;
	    while (pointer <= pointer_stack_top)
	    {
		Block* b = GC_findParentBlock(*static_cast<void**>(*pointer));
		if (b != 0)
		{
		    b->setTag(tag_necessary);
		}
		pointer++;
	    }
	}
	{
	    Block* const * block = begin;
	    unsigned tag = tag_necessary + 1;
	    while (block < end)
	    {
		if ((*block)->getTag() != tag_necessary)
		{
		    (*block)->setTag(tag);
		}
		++block;
	    }
	}
    }


    /*!
       Collect block begins into array for faster access in GC algorithm.
       \return begin and end iterators of obtained array.
     */
    void MemoryPool::GC_prepare(Block** &begin, Block** &end)
    {
	end = (Block**) pointer_stack;
	begin = end;

	Block* block = blocks_end;
	while (block != 0)
	{
	    if (!block->isFree())
	    {
		--begin;
		*begin = block;
	    }
	    block = block->getPrev();
	}
    }


    void MemoryPool::GC_clearMarked(const unsigned tag_necessary)
    { 
	bool need_clean = false;
	do
	{
	    need_clean = false;
	    Block* block = blocks;
	    while(block != 0)
	    {
		assert(block->selfTest(std::cerr));
		if (block->getTag() != tag_necessary)
		{
		    freeBlock(block);
		    break;
		}
		block = block->getNext();
	    }
	} while (need_clean);
    }


    /*!
       Perform universal garbage collection algorithm.
       This algorithm can handle arbitrary pointer graphs (even with cycles).
     */
    void MemoryPool::collectBlockGarbage()
    {
	// total garbage collection
	const unsigned TAG_NECESSARY = 0;
	const unsigned TAG_TOP = 0x0FFFFFFF; // depends on Block::tag capacity
		    
	Block** begin;
	Block** end;   // STL iterator like

	GC_prepare(begin, end);

	GC_setInitialMarks(begin, end, TAG_NECESSARY, TAG_TOP);
	{ // scan blocks
	    Block** block = begin; // block - begin og blocks to be scanned
	    Block** p = begin;     // p - bottom of unknown blocks
	    { // partition
		for (Block** i = begin; i < end; ++i)
		{
		    if ((*i)->getTag() == TAG_NECESSARY)
		    {
			std::swap(*p, *i);
			++p;
		    }
		};
	    }
	    while (block < end && block < p )
	    {
		assert((*block)->getTag() == TAG_NECESSARY);
		assert(!(*block)->isFree());
		GC_extendBlockMark(*block, p, end, TAG_NECESSARY);
		++block;
	    }
	}
	GC_clearMarked(TAG_NECESSARY);
	assert(selfTest(std::cerr));
    }; // MemoryPool::collectBlockGarbage


    /*!
       If pointer == 0 then perform complete garbage collection
       else gaze at the given block only.
     */
    void MemoryPool::collectGarbage(void** pointer)
    {
	if (pointer != 0)
	{
	    Block* block = GC_findParentBlock(*pointer);
	    if (block->notReferenced())
	    {
		freeBlock(block);
	    }
	}
	else
	{
	    collectBlockGarbage();
	}
    };


    /*!
       This class is used for sorting pointers while preparing
       defragmentation.
     */
    class DefragCmp
    {
	public:
	    int operator()(void** const& a, void** const& b) const
	    { return (char*)(*a) < (char*)(*b); }
    };
	    

    /*!
       Prepare defragmentation process by collecting all info about pointers
       into single array.
       We store info about pointers-in-heap in mask, and info about
       pointers-in-porogram-stack in stack-like structure (see PS_*
       routines) so below are two stages of the preparation algorithm.
     */
    void MemoryPool::defragmentPrepare(void*** &begin, void*** &end)
    {
	end = (void***)pointer_stack;
	begin = end; // initially empty set
	unsigned count = 0;
       
	// 1. collect pointers from stack
	{
	    void*** p = end;
	    while (p <= (void***)pointer_stack_top)
	    {
		if (*p != 0)
		{
		    *(-- begin) = *p;
		    count++;
		    if (count > reserve) 
			throw new Error("Inconsistency in data structures:"
					" reserve is not enough (1).");
		}
		p++;
	    }
	}
       
	// 2. collect pointers from mask
	{
	    PointerMask::iterator i(&pointer_mask);
	    while (i.more())
	    {
		if (i.get())
		{
		    void** adr = i.getAdr();
		    *(-- begin) = adr; 
		    count++;
		    if (count > reserve) 
			throw new Error("Inconsistency in data structures:"
					" reserve is not enough(2).");
		}
		++i;
	    }
	}
       
	// 3. sort pointers
	assert(begin <= end);
	std::sort(begin, end, DefragCmp());
	assert((unsigned)abs(end - begin) == count);
	end--;    
    };

    /*!
       Memory defragmentation algorithm
     */
    void MemoryPool::defragment()
    {
	void*** begin;
	void*** end;

	defragmentPrepare(begin, end);

	// p - points to current position in pointers table
	void*** p = begin;
	Block* block = blocks;
	while (block != 0) // scan all blocks and try to shift them
	{
	    Block* next_block = block->getNext(); 
	    if (next_block == 0)
	    {
		break;
	    }
	    if (block->isFree() && !next_block->isFree() && !next_block->isPasted())
	    {
		int delta  = next_block->shift_down_offset();
		assert(delta <= 0);
		if (delta != 0)
		{
		    // save initial range
		    void* bbegin = next_block;
		    void* bend   = (char*)next_block + next_block->getSize() - 1;
		    // first update pointers
		    // because there could be pointers in the block too
		    {
			// update pointers in the stack
			assert(begin <= p);
			while (**p <= bend && p <= end)
			{
			    if (**p >= bbegin && **p != 0)
				(char*)(**p) += delta;
			    ++p;
			}
			void*** f = p;
			while (f <= end) // NOTE this cycle leads to complexity about N^2/2
			{
			    if (bbegin <= *f && *f <= bend)
				(char*)*f += delta;
			    f++;
			}
		    }
		    // now shift
		    next_block->shift_down();
		    pointer_mask.move_range(bbegin, bend, delta);
		}
		next_block = block->getNext();
	    }
	    block = next_block;
	}
	assert(selfTest(std::cerr));
    };


    Block* MemoryPool::findFreeBlock(const uint minSize, 
					    Block* freeBlockHint)
    {
	Block* block = freeBlockHint != 0 ?
			    freeBlockHint : blocks;
	while (block != 0)
	{
	    if (block-> isFree())
		if (block-> getSize() >= minSize)
		    break;
	    block = block-> getNext();
	}
	if (block == 0 && freeBlockHint != 0)
	{
	    block = freeBlockHint;
	    while (block != 0)
	    {
		if (block-> isFree())
		    if (block-> getSize() >= minSize)
			break;
		block = block-> getPrev();
	    }
	}
	return block;
    }


    /*!
       Create new block for utilization.
       \param size (now it) size of new block = header size + body size
       \param initializator Procedure that initializes new block (this function
       		is different for different block types).
     */
    Block* MemoryPool::allocBlock(const uint size, Block::InitializationFunction initializator)
    {
	assert(_initialized);
	assert(initializator != 0);
	Block* newBlock = 0;
	// 1: lookup for appropriate free gap
	Block* block = findFreeBlock(size, freeBlockHint);
	if (block != 0)
	{
	    // 2: insert block
	    newBlock = block->giveBirth(size, initializator);
	}
	if (newBlock == 0)
	{
	    throw new EOutOfMemory();
	    return 0;
	}
	return newBlock;
    };


    void MemoryPool::freeBlock(Block* block)
    {
	assert(_initialized && block != 0);
	assert(!block->isInvalid());
	block->~Block();
    };


    /*!
       Save info about pointer in program stack into memory
       pool's structure.
     */
    void MemoryPool::PS_Push(void** const pointer)
    {
	assert(pointer_stack <= pointer_stack_top + 1);
	if (pointer_stack <= pointer_stack_top) // stack is not empty
	{
	    // first bring in order stack pointers info
	    std::sort(pointer_stack, pointer_stack_top);
	    while (*pointer_stack == 0 && pointer_stack <= pointer_stack_top)
		pointer_stack++;
	}
	// if tail of block list starting to intersect with stack pointers 
	if (!blocks_end->isFree() || pointer_stack-1 <= (void**)(blocks_end+1))
	{
	    optimize();
	    if (!blocks_end->isFree() || available_top <= (void**)(blocks_end+2))
		throw new EOutOfMemory();
	}
	pointer_stack--;
	update_reserve();
	*pointer_stack = pointer;
	assert(pointer_stack <= pointer_stack_top);
    };


    /*!
       Forget about pointer in program stack.
     */
    void MemoryPool::PS_Pop(void** const pointer)
    {
	assert(pointer != 0);
	assert(pointer_stack <= pointer_stack_top);
	while (pointer_stack <= pointer_stack_top && *pointer_stack == 0)
		pointer_stack++;
	
	void** p = pointer_stack;
	if (*p == pointer)
	{
	    *pointer_stack = 0;
	    if (pointer_stack <= pointer_stack_top)
		pointer_stack++;
	}
	else
	{
	    // delete pointer from array
	    while (p <= pointer_stack_top)
	    {
		if (*p == pointer)
		{
		    *p = 0;
		    // squize array
		    void** i = p;
		    while (pointer_stack <= i)
		    {
			if (*i != 0)
			    *p-- = *i;
			i--;
		    }
		    assert(pointer_stack <= pointer_stack_top);
		    pointer_stack = p + 1;
		    return;
		}
		p++;
	    }
	}
	assert(pointer_stack <= pointer_stack_top + 1);
	while (pointer_stack <= pointer_stack_top && pointer_stack == 0)
		pointer_stack++;
    };


    int MemoryPool::PS_Count() const
    {
	int count = 0;
	for (void** p = pointer_stack; p <= pointer_stack_top; ++p)
	    if (*p) count++;
	return count;
    }


    //! before-deletion-callback
    void MemoryPool::on_block_delete(Block* block)
    {
	if (block == blocks_end)
	{
	    blocks_end = blocks_end-> getPrev();
#ifdef NEW_BLOCK_IN_FRONT 
	    freeBlockHint = blocks_end;
#else
	    freeBlockHint = blocks;
#endif	
	}
	block_count--;
	assert(block_count >= 1); // at least one block should always be there 
    };


    //! after-creation-callback
    void MemoryPool::on_block_new(Block* block)
    {
	if (blocks_end->getNext() == block ||
	    blocks_end == block)
	{
	    while (blocks_end->getNext() != 0)
		blocks_end = blocks_end->getNext();
	    assert(blocks_end->getNext() == 0);
#ifdef NEW_BLOCK_IN_FRONT 
	    freeBlockHint = blocks_end;
#else
	    freeBlockHint = blocks;
#endif	
	}
	block_count++;
    };


    void MemoryPool::dumpBlocks(std::ostream& stream,
				       const std::string& prefix) const
    {
	Block* currentBlock = blocks;
	while (currentBlock != 0)
	{
	    currentBlock->dump(stream, prefix);
	    currentBlock = currentBlock->getNext();
	}
    };
	

    void MemoryPool::dumpPointers(std::ostream& stream,
				 const std::string& prefix,
				 const std::string& name) const
    {
	stream << prefix << (name.empty() ? getClassName() : name) << "::";
	Statistics stat(this);
	stream << stat.pointers_in_stack << std::endl;
	for (void*** p = (void***)pointer_stack; p <= (void***)pointer_stack_top; ++p)
	{
	    stream << prefix + "\t" << "s:" << *p << ":" << **p << std::endl;
	}
    };


    void MemoryPool::dump(std::ostream& stream,
				 const std::string& prefix,
				 const std::string& name) const
    {
	stream << prefix << (name.empty() ? getClassName() : name) << "::";
	stream << blocks << ":";
	stream << blocks_end << ":";
	stream << pointer_stack << ":";
	stream << pointer_stack_top << ":";
	stream << pool_top << ":";
	stream << _initialized << ":";
	stream << getCapacity() << ":";
	stream << std::endl;
	dumpBlocks(stream, prefix + "\t");
	dumpPointers(stream, prefix + "\t", "stack pointers");
	pointer_mask.dump(stream, prefix + "\t", "mask pointers");
	stream << std::endl;
    };


    uint MemoryPool::getFreeAmount() const
    {
	Block* block = blocks;
	uint real_free_space = 0;
	while (block != 0)
	{
	    if (block->isFree())
	    {
		real_free_space += block->getCapacity();
	    }
	    block = block->getNext();
	}
	return real_free_space;
    };


    MemoryPool::Statistics::Statistics(const MemoryPool* const memory_pool) :
	pointers(&memory_pool->pointer_mask)
    {
	block_size = sizeof(Block);
	capacity = memory_pool->getCapacity();
	free_space = memory_pool->getFreeAmount();   
	number_of_blocks = 0;
	number_of_used_blocks = 0;
	total_ref_count = 0;
	Block* block = memory_pool->blocks;
	while (block != 0)
	{
	    if (!block->isFree()) 
	    {
		number_of_used_blocks++;
	    }
	    number_of_blocks++;
	    total_ref_count += block->getReferenceCount();
	    block = block->getNext();
	}
	pointers_in_stack = memory_pool->PS_Count();

	PointerMask::Statistics stat(&(memory_pool->pointer_mask));
	pointers_in_mask = stat.pointers_number;
    };		


    /*!
     *
     */
    bool MemoryPool::selfTest(std::ostream& stream) const
    {
	Block* block = blocks;

	assert(blocks <= blocks_end);
	assert((void*)blocks_end < (void*)available_top);
	assert(available_top <= pool_top);
	assert(reserve >= 0);
	assert(pointer_stack <= pointer_stack_top + 1);
	assert(pointer_stack_top < pointer_mask.get_mask_base());    
	// for all registered pointers there should be enough reserved space:
	assert(pointer_mask.get_count() +
		(pointer_stack > pointer_stack_top ?
		 0 : pointer_stack_top - pointer_stack + 1)
		<= reserve
	       );

	if (!((blocks <= blocks_end)
	      &&((void*)blocks_end < (void*)available_top)
	      &&(available_top <= pool_top)
	      &&(reserve >= 0)
	      &&(pointer_stack <= pointer_stack_top + 1)
	      &&(pointer_stack_top < pointer_mask.get_mask_base())    
	      &&(pointer_mask.get_count() +
		    (pointer_stack > pointer_stack_top ?
		     0 : pointer_stack_top - pointer_stack + 1)
		    <= reserve
		   )
	     )
	   )
	    return false;

	if (block->getNext() == 0 && block->getPrev() == 0)
	    if (!block->isFree())
	    {
		assert(false);
		return false;
	    };

	uint real_capacity = 0;
	while (block != 0)
	{
	    if (!block->selfTest(stream))
	    {
		assert(false);
		return false;
	    }
	    real_capacity += block->getSize();
	    if (block->getNext() == 0)
		if (block != blocks_end)
		{
		    assert(false);
		    return false;
		}
	    block = block->getNext();
	}
	if (getCapacity() != real_capacity)
	{
	    assert(false); 
	    return false;
	}
	
	Statistics stat(this);
	if (stat.total_ref_count != stat.pointers_in_stack + stat.pointers_in_mask)
	{
	    assert(false);
	    return false;
	};
	if (stat.total_ref_count != ptr_count)
	{
	    assert(false);
	    return false;
	};
	if (stat.number_of_blocks != block_count)
	{
	    assert(false);
	    return false;
	};
	return true;
    };


    /*!
       Brute force pointer check. 
     */
    bool MemoryPool::isCorrectPointer(const void* pointer) const
    {
	if (pointer == 0)
	{
	    return true;
	}
	Block* block = blocks;
	while (block)
	{
	    if (pointer == block + 1)
	    {
		return true;
	    }
	    block = block->getNext();
	}
	return false;
    };

    
} // namespace dm


// EOF

