/*!
    \file mm/dm.h
    \brief Flood dynamic memory manager interface.
    \author (c) Petr Gladkikh, 2001-2002
    $Id: dm.h,v 1.6 2002/10/20 15:07:32 petr Exp $

    \note Functions with GC prefix are parts of Garbage Collection algorithm.
  
    <h3>Abbreviations</h3>
    <dl>
    <dt>BRC</dt>
    <dd>back reference container. It holds set of references to the block. This block is named "owner" of that container.</dd>
    <dt>GC</dt>
    <dd>garbage collection.</dd>
    </dl>
 */

#ifndef DYN_MEM_CPP_H
#define DYN_MEM_CPP_H
 
#include <iostream>
#include <string>
#include <cassert>
#include "util/util.h"
#include "util/pointer_mask.h"

/*!
   \namespace dm
   \brief Flood's home. 
  
   It contains all Flood's public interfaces and static data members
 */
namespace dm
{

    /*!
	\define NEW_BLOCK_IN_FRONT
	This symbol defines whether barent block will be splitted
	into (new_block, new_unused_block) if defined or into (new_unused_block,
	new_block) if not defined. you can check both cases in extensive
	testing. 
	\note 
	This option should be always defined due to imlemented reservation rules.
	If there no reservation for auxiliary optimization structures then trere 
	is possibility of this choice.
    */
#define NEW_BLOCK_IN_FRONT


    class Error
    {
	public:
	    std::string message;
	    Error(const std::string new_message = "")
	    { 
		message = new_message;
		std::cerr << "ERROR:" << message << std::endl; // DEBUG
	    }
    };

    /*!
       Baisc reference type
     */
    class Reference
    {
	protected:
	    void* data;

	public:
            static std::string getClassName() { return "Reference"; };
	    static void dump(std::ostream& stream,
			     void** to,
			     const std::string& prefix = "",
			     const std::string& name = "");
	    Reference() : data(0) {};
	    Reference(void* new_data) : data(new_data) {};
    };


    class MemoryPool;
    class Block;
    
    /*!
       This class is necessary to collect all properties common for all
       dynamic<T> templates.
     */
    class GenericDynamic : public Reference
    {
	protected:
	    static MemoryPool* memory_pool;
	
	public:
	    static void setPool(MemoryPool* const newMemoryPool);
    };
	

    /*!
       This class represents smart reference of object. It represents "smart pointer"
       functionality.
       All declared user objects should use this class.
     */
    template <class T>
    class dynamic : public GenericDynamic
    {
	private:
	    Block* getBlock() const; //!< returns the block this dynamic data containing
	    void releaseData();

	public:
	    explicit dynamic(const dm::uint length = 0);
	    dynamic(const dynamic& another);
	    ~dynamic();

	    dm::uint getLength() const;
	    T&   operator() (const dm::uint i = 0);
	    T&   operator[] (const dm::uint i) { return (*this)(i); };
	    T&   operator* () { return (*this)[0]; };
	    T*   init(const dm::uint i = 1);
	    T*   ptr() const { return static_cast<T*>(data); };
	    void remove();
	    T*   operator -> () { return ptr(); };

	    dynamic& operator = (const dynamic& other);

	    bool operator == (const dynamic<T>& other) const;
	    bool operator != (const dynamic<T>& other) const;

	    bool operator == (const void* other) const;
	    bool operator != (const void* other) const;
    };


    /*!
       Unit, which contains user information, relations and description.
       A memory pool contains linked list of blocks.
     */
    class Block
    {
	protected:

	    friend class MemoryPool; 

	    template<class T>
	    static Block* t_initInstance(Block* const parent, const dm::uint new_size)
	    {
#ifdef NEW_BLOCK_IN_FRONT	    
		return t_initInstance_2<T>(parent, new_size);
#else
		return t_initInstance_1<T>(parent, new_size);
#endif	    
	    };
	    
	    template<class T>
	    static Block* t_initInstance_1(Block* const parent, const dm::uint new_size);

	    template<class T>
	    static Block* t_initInstance_2(Block* const parent, const dm::uint new_size);

	    static Block* getNewInstanceAddr(Block* const parent, const dm::uint new_size);

	    /*! Says wether small parent will be consumed by it's child.
		If this value is -1 then leave parent whether it is possible.
	    */
	    static const dm::lint CONSUME_TRESHOLD = -1; 	

	    static MemoryPool* memory_pool;
	    
	    Block*   prev; //!< Previous block in chain.
	    Block*   next; //!< Next block in chain.
	    dm::uint size; //!< Size (span) of this block (header_size+body_size) in bytes.
	    
	    //! Packed set of block's properties.
	    struct
	    {
		unsigned used : 1;     //!< If true then this block is allocated for user.
		unsigned invalid : 1;  //!< If true then this is not a block (broken, deleted, etc.) 
		unsigned under_destruction : 1; //!< If true then control is inside destructor. 
		unsigned pasted : 1;   //!< If true then do not relocate the block.
		unsigned tag : 28;     //!< Auxiliary value, used in GC and reordering algorithms.
	    } flags;
	    
            //! Number of references to this block (without those from prev and next blocks in chain).
	    dm::uint reference_counter;
	    
	    void free(const bool clean_pointers = false);	
	    void markAsUsed();
	    void markAsFree();
	    void markAsInvalid();
	    void mergeWithNext();
	    Block(Block& parent); //!< \note Should be called using "new ptr ctor;" construction only

	    dm::uint getBRCount() const;

	    // reordering
	    int shift_down_offset() const;
	    dm::uint shift_down();
	    void operator= (const Block& another_block);

	private :
	    void initReferences();

	public:
	    /*!
	       This typedef defines form of block initialization function.
	       \param parent Parent for the instance, where latter should be located.
	       \param size   Size of new block instance.
	     */
	    typedef Block* InitializationFunction(Block* const parent,    
						  const dm::uint size);
	    static Block* initInstance(Block* const parent,
				       const dm::uint new_size);   //!< wrapper for constructor
	    virtual InitializationFunction* getInitializator() const
		{ return initInstance; };
	    
	    static void setPool(MemoryPool* const newMemoryPool);
	    Block(const dm::uint new_size,
			 Block* const new_prev = 0,
			 Block* const new_next = 0); //!< Bootstrap constructor.
	    virtual  ~Block();
	    virtual  dm::uint getCapacity() const;
	    virtual  dm::uint getHeaderSize() const;
	    Block*   resize(dm::uint new_size);
	    bool     addressInside(void* const addr) const;
	    bool     addressInsideBody(void* const addr) const;
	    virtual  std::string getClassName() const { return "Block"; };
	    dm::uint getSize() const;
	    void*    getBody() const { return (void*)((char*)this + getHeaderSize()); };
	    Block*   getNext() const;
	    Block*   getPrev() const;
	    bool     isFree() const;
	    bool     isInvalid() const;
	    virtual  bool isEmpty() const { return false; };
	    bool     isPasted() const { return flags.pasted; };
	    void     setPasted(const bool new_pasted) { flags.pasted = new_pasted; };
	    Block*   giveBirth(const dm::uint sizeToUse,
			      InitializationFunction dad);
	    virtual  bool registerPointer(void** value);
	    bool     forgetPointer(void** value);
	    bool     notReferenced() const;
	    dm::uint getReferenceCount() const;
	    void     setTag(const unsigned new_tag) { flags.tag = new_tag; };
	    unsigned getTag() const { return flags.tag; };
	    virtual  void dump(std::ostream& stream,
			      const std::string& prefix = "",
			      const std::string& name = "");
	    bool selfTest(std::ostream& stream) const;
    }; // end of class Block


    /*!
       This class knows about class instances in own body
       and call their constructors and destructors when appropriate.
     */
    template <class T>
    class SmartBlock : public Block
    {
	protected:
	    static const dm::uint ELEMENT_SIZE = sizeof(T);

	public:
	    SmartBlock(Block& parent);
	    static Block* initInstance(Block* const parent,
				       const dm::uint new_size); // wrapper for constructor
	    virtual InitializationFunction* getInitializator() const
		{ return initInstance; };

	    T*       getBody();
	    dm::uint getLength();
	    T&       operator[](const dm::uint index);
	    virtual  ~SmartBlock();
    };


    /*!
       Heap manager.
     */
    class MemoryPool
    {
	protected:

	    friend class Block;
	    
	    bool _initialized; //!< Initialization was already performed if true. 
	    Block *blocks;     //!< First block in chain.
	    Block *blocks_end; //!< Last block in chain.
	    void** pool_top;   //!< Max (upper) address available for DMP.
	    void** available_top; //!< Max address available for blocks allocation.
	    dm::uint reserve;     //!< Reserved amount for optimization algorithms.
	    /*
		assertions: blocks <= blocks_end < available_top <= pool_top
			    reserve >= 0
	     */
	    
	    dm::uint capacity; //!< Size of whole address range available 
	    Block* findFreeBlock(const dm::uint minCapacity, 
				 Block* freeBlockHint = 0);
	    void freeBlock(Block* block);
	    void collectBlockGarbage();
	    Block* freeBlockHint;
	    dm::uint block_count;
	    dm::uint ptr_count;

	    dm::uint calculate_reserve() const;
	    void update_reserve();

	    PointerMask pointer_mask;
	    
	    void** pointer_stack;
	    void** pointer_stack_top;
	    /*
		assertion pointer_stack <= pointer_stack_top
	     */
	    void PS_Push(void** const pointer);
	    void PS_Pop(void** const pointer);
	    int  PS_Count() const;
	    
	    // Notification handlers
	    void on_block_delete(Block* block);
	    void on_block_new(Block* block);

	private:
	    void GC_extendBlockMark(Block* const block,
				    Block** &begin,
				    Block** &end,
				    const dm::uint tag_necessary);
	    Block* GC_findParentBlock(const void* pointer);
	    void GC_setInitialMarks(Block* const * begin,
				    Block* const * end,
				    const unsigned tag_necessary,
				    const unsigned tag_top);
	    void GC_prepare(Block** &begin, Block** &end);
	    void GC_clearMarked(const unsigned tag_necessary);
	    void defragmentPrepare(void*** &begin, void*** &end);

	public:
	    bool isInitialized() const { return _initialized; };

	    class Parameters
	    {
		public:
		    Parameters() : capacity(0) {};
		    dm::uint capacity;
	    };

	    class Statistics
	    {
		public:
		    Statistics(const MemoryPool* const memory_pool);
		    dm::uint block_size;
		    dm::uint capacity;
		    dm::uint free_space;
		    dm::uint number_of_blocks;
		    dm::uint number_of_used_blocks;
		    dm::uint pointers_in_stack;
		    dm::uint pointers_in_mask;
		    PointerMask::Statistics pointers;
		    dm::uint total_ref_count;
	    };

	    class EOutOfMemory : public Error {};

	    MemoryPool(const dm::uint initialCapacity = 0);
	    virtual  ~MemoryPool();
	    bool     init(const Parameters parameters);
	    void*    getPoolBottom() const { return blocks; };
	    void*    getPoolTop() const { return (char*)blocks + capacity; };
	    dm::uint getFreeAmount() const;
	    dm::uint getCapacity() const;
	    virtual  std::string getClassName() const { return "MemoryPool"; };
	    void     registerPointer(void** pointer);
	    void     forgetPointer(void** pointer);
	    bool     optimize();
	    void     collectGarbage(void** pointer = 0);
	    void	 defragment();
	    Block*   allocBlock(const dm::uint size, Block::InitializationFunction initializator);
	    void*	 alloc(const dm::uint size);
	    void	 free(void** blockToDelete);
	    void	 dumpBlocks(std::ostream& stream, const std::string& prefix = "") const;
	    void	 dumpPointers(std::ostream& stream, const std::string& prefix = "", const std::string& name = "") const;
	    void	 dump(std::ostream& stream, const std::string& prefix = "", const std::string& name = "") const;
	    bool	 selfTest(std::ostream& stream) const;
	    bool	 isCorrectPointer(const void* pointer) const;
    };


    // ========================================================
    // Inline methods implementation


    // --------------------------------------------------------
    // class dynamic


    template<class T>
    inline dynamic<T>::dynamic(const dm::uint length)
    {
	if (length > 0)
	{
	    init(length);
	}
    };


    template<class T>
    inline dynamic<T>::dynamic(const dynamic& another)
    {
	assert(data == 0);
	*this = another;
    };


    template <class T>
    inline T* dynamic<T>::init(const dm::uint length)
    {
	if (data != 0)
	{
	    memory_pool->forgetPointer(&data);
	}
	assert(data == 0);
	Block* block = memory_pool->allocBlock(sizeof(SmartBlock<T>) + length*sizeof(T),
					       SmartBlock<T>::initInstance);
	assert(block != 0);
	data = block->getBody();    
	assert(data != 0);
	memory_pool->registerPointer(&data);
	return static_cast<T*>(data);
    };


    template <class T>
    inline dynamic<T>::~dynamic() 
    {
	remove(); 
    };


    template <class T>
    inline dm::uint dynamic<T>::getLength() const
    {
	return getBlock()->getLength();
    };


    template <class T>
    inline T& dynamic<T>::operator() (const dm::uint i)
    {
	assert(data != 0);
	return static_cast<T*>(data)[i];
    };


    template <class T>
    inline void dynamic<T>::remove()
    {
	if(data != 0)
	{
	    releaseData(); 
	}
    };


    template <class T>
    inline Block* dynamic<T>::getBlock() const
    {
	assert(data != 0);
	return static_cast<Block*>(data) - 1;
    }


    template <class T>
    inline void dynamic<T>::releaseData()
    {
	memory_pool->forgetPointer(&data);
    }


    template <class T>
    inline dynamic<T>& dynamic<T>::operator = (const dynamic<T>& other)
    {
	if (*this == other)
	{
	    return *this;
	}
	// this handle will protect data from deletion until end of this procedure
	dynamic<T> temporary_handle;
	if (data != 0)
	{
	    temporary_handle = *this;
	    releaseData();
	}
	if (other.data != 0)
	{
	    data = other.data;
	    memory_pool->registerPointer(&data);
	}
	else
	{
	    data = 0;
	}
	return *this;
    };


    template <class t>
    inline bool dynamic<t>::operator == (const dynamic<t>& other) const
    {
	return this->data == other.data;
    };


    template <class t>
    inline bool dynamic<t>::operator != (const dynamic<t>& other) const
    {
	return ! ((*this) == other);
    };


    template <class t>
    inline bool dynamic<t>::operator == (const void* other) const
    {
	return this->data == other;
    };


    template <class t>
    inline bool dynamic<t>::operator != (const void* other) const
    {
	return ! ((*this) == other);
    };


    // --------------------------------------------------------
    // class Block


    inline dm::uint Block::getSize() const
    {
	if (next == 0)
	    return (char*)(memory_pool->pointer_stack) - (char*)this;
	else
	    return size;
    };


    inline bool Block::addressInside(void* const addr) const
    {
	return (this <= addr) 
		&& (addr < reinterpret_cast<const char*>(this) + getSize());
    };


    inline bool Block::addressInsideBody(void* const addr) const
    {
	return (this+1 <= addr) 
		&& (addr < reinterpret_cast<const char*>(this) + getSize());
    };


    inline Block* Block::getNext() const
    {
	return next;
    };


    inline Block* Block::getPrev() const
    {
	return prev;
    };


    inline bool Block::isFree() const
    {
	return !flags.used;
    };


    inline bool Block::isInvalid() const
    {
	return flags.invalid;
    };


    inline void Block::markAsUsed()
    {
	 flags.used = true;
    };


    inline void Block::markAsFree()
    {
	flags.used = false;
    };


    inline void Block::markAsInvalid()
    {
	flags.invalid = true;
    };


    /*!
       We can not obtain pointer to constructor (due to ISO C++
       language restrictions) thus we need this wrapper.
       \par
       Split parent block into two. Latter one will be new block
       (see also NEW_BLOCK_IN_FRONT). Parent become smaller or is 
       consumed by child.
    */
    template<class T>
    Block* Block::t_initInstance_1(Block* const parent,
				   const dm::uint new_size)
    {
	return dynamic_cast<Block*> (
		   new ((void*)getNewInstanceAddr(parent, new_size)) T(*parent)
	       );
    };


    /*!
        Differene between t_initInstance_1 and this function:
	Insert new block instead of parent. Tail become new empty
	block.
    */
    template<class T>
    Block* Block::t_initInstance_2(Block* const parent,
				   const dm::uint new_size)
    {
	Block* new_instance = getNewInstanceAddr(parent,
						 parent->getSize() - new_size );
	if (new_instance >= parent)
	{
	    new ((void*)new_instance) Block(*parent);
	}
	return dynamic_cast<Block*> (
		   new ((void*)parent) T(*parent)
	       );
    };


    // --------------------------------------------------------
    // class SmartBlock


    template<class T>
    Block* SmartBlock<T>::initInstance(Block* const parent, const dm::uint new_size)
    {
	return t_initInstance<SmartBlock>(parent, new_size);
    };


    /*! 
	\todo FIXIT: If block is larger than expected minimal
	array size + ELEMENT_SIZE - 1
	this function will return value greater than expected :{ 
    */
    template<class T>
    dm::uint SmartBlock<T>::getLength()
    {
	return getCapacity() / ELEMENT_SIZE; 
    }


    template<class T>
    T* SmartBlock<T>::getBody()
    {
	return (T*)Block::getBody();
    };


    /*! \todo See comments to SmartBlock::getLength
     */
    template<class T>
    T& SmartBlock<T>::operator[](const dm::uint index)
    {
	assert(index < getLength());
	return getBody()[index]; 
    }


    template<class T>
    SmartBlock<T>::SmartBlock(Block& parent)
	: Block(parent)
    {
	// initialize array of class instanses in the block
	for (dm::uint count = 0; count < getLength(); count++)
	{
	    new ((void*)(getBody() + count)) T();
	};
    };


    template<class T>
    SmartBlock<T>::~SmartBlock()
    {
	if (flags.under_destruction)
	    return;
	flags.under_destruction = true;
	// destroy array of class instances in the block
	for (dm::uint count = 0; count < getLength(); count++)
	{
	    getBody()[count].~T();
	};
	flags.under_destruction = false;
    };


    // --------------------------------------------------------
    // class MemoryPool


    inline void* MemoryPool::alloc(const dm::uint size)
    {
	assert(_initialized);
	Block* newBlock = allocBlock(sizeof(Block) + size, Block::initInstance);
	return newBlock->getBody();
    };


    inline void MemoryPool::free(void** blockToDelete)
    {
	assert(_initialized && blockToDelete != 0 && *blockToDelete != 0);
	freeBlock(*(Block**)blockToDelete - 1);
	*blockToDelete = 0;
    };


    inline void MemoryPool::registerPointer(void** pointer)
    {
	assert(_initialized && pointer != 0 && *pointer != 0);
	Block* block = (static_cast<Block*>(*pointer)-1);
	assert(block->selfTest(std::cerr));
	block->registerPointer(pointer);
	if (pointer > pool_top) // suppose pointer is in stack
	{
	    PS_Push(pointer);
	}
	else
	{
	    pointer_mask.registerPointer(pointer);
	}
	ptr_count++;
	update_reserve();
    };


    inline void MemoryPool::forgetPointer(void** pointer)
    {    
	assert(_initialized && pointer != 0 && *pointer != 0);
	Block* block = GC_findParentBlock(*pointer);
	if (block != 0 && !block->isFree())
	{
	    assert(block->selfTest(std::cerr));
	    block->forgetPointer(pointer);
	}
       
	if (pointer > pool_top) // suppose it is in stack
	{
	    PS_Pop(pointer);
	}
	else
	{
	    pointer_mask.forgetPointer(pointer);
	}
	*pointer = 0;
	ptr_count--;
	update_reserve();
    };


    inline dm::uint MemoryPool::getCapacity() const
    {
	assert(_initialized);
	return (char*)pointer_stack - (char*)blocks;
    };

} // namespace dm

#endif
// EOF dynamic memory manager interface

