/*!
    \file flat_tree.h
    \brief A binary tree implementation.
    \author Petr Gladkikh, 2001    
    $Id: flat_tree.h,v 1.2 2002/09/17 17:44:14 petr Exp $
    
    This file contains interface to binary tree which do not
    need memory management. 
*/

#ifndef DM_PBTREE_DECL
#define DM_PBTREE_DECL

#include "util/util.h"
#include <cassert>
#include <ctype.h>

namespace dm
{
    //! Comparator function type
    typedef int (*Comparator)(const void*, const void*);

    /*!
     * \brief Flat binary tree
     *
     * This implementaion of binary tree does not use any dynamic
     * allocation facilities. Flat tree represents binary tree
     * using simplest memory allocation technique and only data
     * structure it requires is continuous one dimansional array. 
     * \par
     * This class was considered as candidate for pointer registry.
     */
    template <class T>
    class FBTree
    {
      protected:
	static const dm::uint NULL_SLOT  = (dm::uint)(-1);
	static const dm::uint EMPTY_SLOT = (dm::uint)(-2);

	class FBTreeNode
	{
	  protected:
	    dm::uint left;
	    dm::uint right;

	  public:
	    T element;
	    FBTreeNode()   { init(); };
	    void init()    { left = EMPTY_SLOT; right = NULL_SLOT; };
	    void setLeft(const dm::uint index) { left = index; };
	    dm::uint getLeft()  { return left; };
	    void setRight(const dm::uint index) { right = index; };
	    dm::uint getRight() { return left; };
	    bool isEmpty() { return left == EMPTY_SLOT; };

	    void recycle(const dm::uint prevRecycled);
	    dm::uint  use();
	};

	dm::uint size; 
	FBTreeNode* body;
	dm::uint root;
	dm::uint freeList;
	dm::Comparator comparator;
	void recycle(const dm::uint index);
	dm::uint use();
	FBTreeNode& get(const dm::uint index) const;

      public:
	static dm::uint getFitSize(const dm::uint size);
	FBTree(void* const newBody, const dm::uint newSize,
	       dm::Comparator newComparator);
	~FBTree();    
	dm::uint getCapacity() { return size; }
	void insert(const T& element);
	void remove(const T& element);

	class Iterator
	{
	  protected:
	    dm::uint current;
	    const FBTree* tree;

	  public:
	    Iterator(const FBTree* container) { current = 0; tree = container; };
	    void goRoot();
	    bool hasNext();
	    T*   getNext();
	};
    };


    // --------------------------------------------------------
    // class FBTreeNode


    template <class T>
    inline void  FBTree<T>::FBTreeNode::recycle(const dm::uint prevRecycled)
    {
      init();
      right = prevRecycled;
    };


    template <class T>
    inline dm::uint FBTree<T>::FBTreeNode::use()
    {
      assert(left == EMPTY_SLOT);
      dm::uint next = right;
      left = NULL_SLOT;
      right = NULL_SLOT;
      return next;
    };


    // --------------------------------------------------------
    // class FBTree


    template <class T>
    inline dm::uint FBTree<T>::getFitSize(const dm::uint size)
    {
      return size*sizeof(FBTreeNode);
    };


    template <class T>
    inline FBTree<T>::FBTree(void* const newBody, const dm::uint newSize,
			     dm::Comparator newComparator) 
    {
      assert(newBody != 0 && 0 < newSize);
      size = newSize; 
      comparator = newComparator;
      body = (FBTreeNode*)newBody;
      root = NULL_SLOT;
      freeList = 0;
      body[0].init();
      // recycle all
      for (dm::uint i = 1; i < newSize; i++)
      {
	recycle(i);
      }
    };


    template <class T>
    inline void FBTree<T>::recycle(const dm::uint index)
    {
      body[index].recycle(freeList); 
      freeList = index;
    };


    template <class T>
    inline dm::uint FBTree<T>::use()
    {
      dm::uint newNode = freeList;
      assert(newNode != NULL_SLOT); // no other overflow checking !!
      freeList = body[freeList].use();
      return newNode; 
    };


    template <class T>
    inline typename FBTree<T>::FBTreeNode& FBTree<T>::get(const dm::uint index) const
    {
      return body[index];
    };


    template <class T>
    inline void FBTree<T>::insert(const T& element)
    {
      dm::uint node = root;
      dm::uint prev_node = node;
      int cmp_result = 0;
      while (node != NULL_SLOT)
      {    
	cmp_result = comparator(&element, &get(node).element);
	if (cmp_result == 0)
	{
	  // already in this tree
	  return;
	}
	prev_node = node;
	if (cmp_result > 0) 
	{
	  node = get(node).getRight();
	}
	else 
	{
	  node = get(node).getLeft();
	}
      }
      dm::uint new_node = use(); 
      if (root == NULL_SLOT)
      {
	root = new_node;
      }
      if (prev_node != NULL_SLOT)
      {
	if (cmp_result > 0)
	{
	  get(prev_node).setRight(new_node); 
	}
	else
	{
	  get(prev_node).setLeft(new_node); 
	}
      }
      get(new_node).element = element;
    };


    template <class T>
    inline void FBTree<T>::remove(const T& element)
    {
      dm::uint node = root;
      dm::uint prev_node = node;
      enum {left_child, right_child, is_root} child_kind = is_root;
      int cmp_result;
      while (node != NULL_SLOT)
      {    
	cmp_result = comparator(&element, &get(node).element);
	if (cmp_result == 0)
	{
	  switch (child_kind)
	  {
	    case left_child:  get(prev_node).setLeft(NULL_SLOT);  break;
	    case right_child: get(prev_node).setRight(NULL_SLOT); break;
	    case is_root:;
	  };
	  recycle(node); 
	  return;
	}
	prev_node = node;
	if (cmp_result > 0) 
	{
	  child_kind = right_child;
	  node = get(node).getRight();
	}
	else 
	{
	  child_kind = left_child;
	  node = get(node).getLeft();
	}
      }
    };


    template <class T>
    inline FBTree<T>::~FBTree()
    {
      body = 0;
      size = 0; 
    };    


    template <class T>
    void FBTree<T>::Iterator::goRoot()
    {
      current = 0;
    };


    template <class T>
    bool FBTree<T>::Iterator::hasNext()
    {
      while (current < tree->size && tree->get(current).isEmpty())
      {
	current++;
      };
      return current < tree->size;
    };


    template <class T>
    T* FBTree<T>::Iterator::getNext()
    {
      if (hasNext())
      {
	return &(tree->get(current++).element);
      } 
      return 0;
    };
    
} // namespace dm


#endif
