/*!
    \file pointer_mask.cc
    \brief Pointer set based on bit mask
    \author (c) Petr Gladkikh, 2001
    $Id: pointer_mask.cc,v 1.3 2002/10/20 15:19:45 petr Exp $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Library General Public License as published
    by the Free Software Foundation; either version 2, or (at your option) any
    later version.
*/

#include "pointer_mask.h"
#include <algorithm>
#include <iostream> // DEBUG


// --------------------------------------------------------
//  class PointerMask


/*
   Initialize pointer mask by placing it at the top of specified pool and
   return new pool top 
 */
void** PointerMask::init(void** const new_pool_base,
			 void** const new_pool_top)
{
    assert(new_pool_base < new_pool_top);
    size = (new_pool_top - new_pool_base) / (BITS_PER_WORD + 1) + 1;
    base = (MaskWord*)(new_pool_top - size + 1); 
    assert((void**)base > new_pool_base);
    pool_base = new_pool_base;
    pool_top = base - 1;
    assert(pool_base <= pool_top);
    clean();
    return (void**)base;
};


/*
 * remove all registrations, empty the set
 */
inline void PointerMask::clean()
{
    assert(base != 0 && size > 0);
    std::fill(base, base + size, 0);
}

/*
 * get number of pointers in this mask
 */
unsigned PointerMask::get_count() const
{
    // simplest counting algorithm
    unsigned pointers_number = 0;
    MaskWord* p = base;
    dm::uint count = size;
    while(count > 0)
    {
	if (*p != 0)
	{
	    MaskWord mask = 1;
	    while (mask != 0)
	    {
		if (mask & *p)
		    ++pointers_number;
		mask <<= 1;
	    }
	}
	--count;
	++p;
    }
/*
    PointerMask::iterator i(this);
    while (i.more())
    {
	if (i.get())
	    pointers_number++;
	++i;
    }
*/    
    return pointers_number;
}

// --------------------------------------------------------
// class PointerMask::iterator


PointerMask::iterator::iterator(const PointerMask* const new_owner,
				const void* const pointer)
{
    assert(new_owner != 0);
    owner = new_owner;
    unsigned index;
    if (pointer != 0)
    {
	assert (owner->pool_base <= pointer && pointer <= owner->pool_top);
	index = (void**)pointer - (void**)owner->pool_base;
    }
    else
	index = 0;

    this_word = owner->base + (index / BITS_PER_WORD);
    bit = index % BITS_PER_WORD;

    if (!get() && pointer == 0)
	++(*this);
};


PointerMask::iterator::iterator PointerMask::iterator::operator++ () 
{
    do
    {
	if ((bit == BITS_PER_WORD - 1) && more())
	    do
	    {
		this_word++;
	    } while (more() && (*this_word == 0));

	bit = (bit + 1) % BITS_PER_WORD;

    } while (more() && !get());
   
    return *this;
};


/*
 * add delta to registered pointers within given range 
 * begin <= ... <= end
 */
void PointerMask::update_pointers(const void* begin,
				  const void* end,
				  const int delta)
{
    iterator i(this, begin);
    void** adr = i.getAdr();
    while (i.more() && adr <= end)
    {
	adr = i.getAdr();
	if (begin <= adr && adr < end && i.get())
	    (char*)(*adr) += delta;
	++i;
    }
};


/*
 * update information about moved pointers  
 */
void PointerMask::move_range(void* begin,
			     void* end,
			     const int delta)
{
    // TODO - improve performance
    if (delta == 0 || end < begin)
	return;
    
    assert(delta < 0);
    for (char* i = static_cast<char*>(begin); i <= static_cast<char*>(end); ++i)
    {
	iterator src(this, i);
	iterator dst(this, static_cast<char*>(i)+delta);
	if (src.get())
	{
	    dst.set(true);
	    src.set(false);
	}
    }
}; 


void PointerMask::dump(std::ostream& stream,
		       const std::string& prefix,
		       const std::string& name) const
{
    Statistics stat(this);
    stream << prefix << name << "::" 
	   << base << ":"
	   << size << ":"
	   << stat.pointers_number
	   << std::endl;
    iterator i(this);
    while (i.more())
    {
	stream << prefix << "\tm:" << i.getAdr() << ":" << *(i.getAdr()) << std::endl;
	++i;
    }
};


// --------------------------------------------------------
// class PointerMask::Statistics


PointerMask::Statistics::Statistics(const PointerMask* const pm)
{
    capacity = pm->size * pm->BITS_PER_WORD;
    pointers_number = pm->get_count();
};


// EOF pointer mask

