/*!
    \file pointer_mask.h
    \brief Pointer container based on bit mask
    \author (c) Petr Gladkikh, 2001
    $Id: pointer_mask.h,v 1.2 2002/10/20 15:07:32 petr Exp $
*/ 

#ifndef DM_POINTER_MASK_DECL
#define DM_POINTER_MASK_DECL

#include "util/util.h"
#include <cassert>
#include <string>
#include <iostream>

class PointerMask
{
  protected:
    typedef unsigned MaskWord; 
    static const unsigned BITS_PER_WORD = sizeof(MaskWord) * 8;
    MaskWord* base; // mask base (lowest address)
    dm::uint  size; // mask size in MaskWords
    void*     pool_base; // pool base where pointers are located (lowest address of the pool)
    void*     pool_top;  // pointers to watch on are located between pool_base
                         // and pool_top (including both)
   
    MaskWord* getMaskWordAddr(const void* pointer);
    MaskWord  getWordMask(const void* pointer);

    void clean();
   
  public:
   
    class iterator
    {
	friend class PointerMask;

	protected:
	    const PointerMask* owner;
	    MaskWord* this_word; 
	    unsigned  bit; 
	    MaskWord get_mask() const;
	   
	public:
	    iterator(const PointerMask* const new_owner, const void* const pointer = 0);
	    iterator operator++ (); // get next geristered pointer
	    bool more() const;
	    bool get() const;
	    void** getAdr() const;
	    void set(const bool new_value);
    };
    
    class Statistics
    {
	public:
	    Statistics(const PointerMask* const pm);
	    unsigned capacity;
	    unsigned pointers_number;
    };
   
    unsigned get_count() const;
    void** get_mask_base() const
    	{ return (void**)base; };
    
    void** init(void** const new_pool_base,
		void** const new_pool_top);
    void registerPointer(const void* const pointer);
    bool isRegistered(const void* const pointer) const;
    void forgetPointer(const void* const pointer);
    void update_pointers(const void* begin,
			 const void* bend,
			 const int delta);
    void move_range(void* begin,
		     void* end,
		     const int delta); 
    void dump(std::ostream& stream,
	      const std::string& prefix,
	      const std::string& name ) const;
};


// --------------------------------------------------------
// class PointerMask::iterator


inline bool PointerMask::iterator::more() const
{
    return this_word < owner->base + owner->size;
};


inline PointerMask::MaskWord PointerMask::iterator::get_mask() const
{
    return (MaskWord)1 << bit;
}


inline bool PointerMask::iterator::get() const
{
    return (*this_word) & get_mask();
};


inline void** PointerMask::iterator::getAdr() const
{
    return (void**)(owner->pool_base) + (this_word - owner->base)*BITS_PER_WORD + bit;
};


inline void PointerMask::iterator::set(const bool new_value)
{
    if (new_value)
    {
	*this_word |= get_mask();
    }
    else
    {
	*this_word &= ~get_mask();
    }
};


// --------------------------------------------------------
// class PointerMask


inline void PointerMask::registerPointer(const void* const pointer)
{
    iterator i(this, pointer);
    i.set(true);
    assert(i.get());
};


inline bool PointerMask::isRegistered(const void* const pointer) const
{
    iterator i(this, pointer);
    return i.get();
};


inline void PointerMask::forgetPointer(const void* const pointer)
{
    iterator i(this, pointer);
    i.set(false);
    assert(!i.get());
};


#endif

